import os
try:
    from . import utils
except ImportError:
    import sys
    sys.path.append(os.path.dirname(__file__))
    import utils
import traceback

@utils.cors('POST')
def quotas(request):

    logger = utils.CloudLogger().log_request(request)
    try:

        # 1) Validate the token


        token = utils.extract_token(request.headers, None)
        if token is None:
            return (
                {
                    'error': 'Bad Request',
                    'message': 'Token must be provided in header or body'
                },
                400
            )

        token_data = utils.get_token_info(token)
        if 'error' in token_data:
            return (
                {
                    'error': 'Invalid Token',
                    'message': token_data['error_description'] if 'error_description' in token_data else 'Google rejected the client token'
                },
                401
            )

        # 2) Check service account
        default_session = utils.generate_default_session(scopes=['https://www.googleapis.com/auth/cloud-platform'])
        account_email = utils.ld_acct_in_project(token_data['email'])
        response = utils.query_service_account(default_session, account_email)
        if response.status_code >= 400:
            return (
                {
                    'error': 'Unable to query service account',
                    'message': response.text
                },
                400
            )
        if response.json()['email'] != account_email:
            return (
                {
                    'error': 'Service account email did not match expected value',
                    'message': response.json()['email'] + ' != ' + account_email
                },
                400
            )

        # 3) Query quota usage
        project_usage = default_session.get(
            'https://www.googleapis.com/compute/v1/projects/{project}'.format(
                project=os.environ.get('GCP_PROJECT')
            )
        )
        if project_usage.status_code != 200:
            return (
                {
                    'error': 'Invalid response from Google',
                    'message': '(%d) : %s' % (
                        project_usage.status_code,
                        project_usage.text
                    )
                },
                400
            )
        quotas = [
            {
                **quota,
                **{
                    'percent':  ('%0.2f%%' % (100 * quota['usage'] / quota['limit'])) if quota['limit'] > 0 else '0.00%'
                }
            }
            for quota in project_usage.json()['quotas']
        ]
        for region_name in utils.enabled_regions():
            region_usage = default_session.get(
                'https://www.googleapis.com/compute/v1/projects/{project}/regions/{region}'.format(
                    project=os.environ.get('GCP_PROJECT'),
                    region=region_name
                )
            )
            if region_usage.status_code != 200:
                return (
                    {
                        'error': 'Invalid response from Google',
                        'message': '(%d) : %s' % (
                            region_usage.status_code,
                            region_usage.text
                        )
                    },
                    400
                    )
            quotas += [
                {
                    **quota,
                    **{
                        'percent':  ('%0.2f%%' % (100 * quota['usage'] / quota['limit'])) if quota['limit'] > 0 else '0.00%',
                        'metric': region_name+'.'+quota['metric']
                    }
                }
                for quota in region_usage.json()['quotas']
            ]
        return (
            {
                'raw': quotas,
                'alerts': [quota for quota in quotas if quota['limit'] > 0 and quota['usage']/quota['limit'] >= 0.5]
            },
            200
        )
    except:
        logger.log_exception()
        return (
            {
                'error': 'Unknown Error',
                'message': traceback.format_exc()
            },
            500
        )



import os
try:
    from . import utils
except ImportError:
    import sys
    sys.path.append(os.path.dirname(__file__))
    import utils
import base64
import hmac
import requests
import json
import time
import traceback

@utils.cors("POST")
def update(request):
    """
    Handles update request from the master update webhook
    """
    logger = utils.CloudLogger().log_request(request)
    try:
        # 1) Validate the request
        if 'X-Lapdog-Signature' not in request.headers:
            return {
                'error': 'Missing Signature',
                'message': "The required X-Lapdog-Signature header was not provided"
            }, 400

        signature = request.headers['X-Lapdog-Signature']
        data = request.get_json()
        if not isinstance(data, dict):
            return (
                {
                    'error': "Bad Request",
                    'message': ("No data was provided" if data is None else "Expected JSON dictionary in request body")
                },
                400
            )
        result = utils.verify_signature(
            bytes.fromhex(signature),
            json.dumps(data).encode(),
            utils.UPDATE_KEY_PATH,
            _is_blob=False
        )
        if not result:
            return {
                'error': 'Bad Signature',
                'message': "The provided signature for this update was invalid"
            }, 401
        if not ('tag' in data and 'url' in data and 'random' in data):
            return {
                'error': 'Missing parameters',
                'message': 'Missing one or more of the required parameters "tag", "url", and "random"',
            }, 400

        # 2) Build pipeline to boot update VM

        regions = utils.enabled_regions()
        if len(regions) < 1:
            return {
                'error': 'No regions',
                'message': 'There are no regions enabled in this project'
            }, 503

        pipeline = {
            'pipeline': {
                'actions': [
                    {
                        'imageUri': 'gcr.io/broad-cga-aarong-gtex/self_update:' + utils.UPDATE_IMAGE_TAG,
                        'commands': ['/update.sh'],
                        'environment': {
                            'LAPDOG_PROJECT': os.environ.get('GCP_PROJECT'),
                            'LAPDOG_LOG_PATH': "gs://{bucket}/update-logs/{time}-{tag}/".format(
                                bucket=utils.ld_meta_bucket_for_project(),
                                tag=data['tag'],
                                time=int(time.time())
                            ),
                            'LAPDOG_CLONE_URL': data['url'],
                            'LAPDOG_NAMESPACE': utils.getblob(
                                'gs://{bucket}/resolution'.format(
                                    bucket=utils.ld_meta_bucket_for_project(os.environ.get('GCP_PROJECT'))
                                )
                            ).download_as_string().decode(),
                            'LAPDOG_TAG': data['tag']
                        }
                    }
                ],
                'resources': {
                    'projectId': os.environ.get("GCP_PROJECT"),
                    'regions': regions,
                    'virtualMachine': {
                        'machineType': 'f1-micro',
                        'preemptible': False,
                        'labels': {
                            'lapdog-execution-role': 'self-update',
                        },
                        'serviceAccount': {
                            'email': 'lapdog-update@{}.iam.gserviceaccount.com'.format(os.environ.get('GCP_PROJECT')),
                            'scopes': [
                                "https://www.googleapis.com/auth/cloud-platform",
                                "https://www.googleapis.com/auth/compute",
                                "https://www.googleapis.com/auth/devstorage.read_write",
                                "https://www.googleapis.com/auth/genomics"
                            ]
                        },
                        'bootDiskSizeGb': 20,
                        'network': {
                            'name': 'default',
                            'usePrivateAddress': False
                        }
                    }
                },
            }
        }

        # 3) Launch pipeline

        logger.log(
            "Launching PAPIv2 pipeline",
            pipeline=pipeline['pipeline'],
            severity='NOTICE'
        )
        response = utils.generate_default_session(
            [
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/compute",
                "https://www.googleapis.com/auth/genomics"
            ]
        ).post(
            'https://genomics.googleapis.com/v2alpha1/pipelines:run',
            headers={
                'Content-Type': 'application/json'
            },
            json=pipeline
        )
        try:
            if response.status_code == 200:
                return response.json()['name'], 200
            return (
                {
                    'error': 'Unable to start update',
                    'message': 'Google rejected the pipeline request (%d) : %s' % (response.status_code, response.text)
                },
                400
            )
        except:
            logger.log_exception("PAPIv2 request failed")
            return (
                {
                    'error': 'Unable to start update',
                    'message': traceback.format_exc()
                },
                500
            )
    except:
        logger.log_exception()
        return {
            'error': 'Unknown error',
            'message': traceback.format_exc()
        }, 500

@utils.cors("POST")
def webhook(request):
    """
    Admins: Use this to trigger a self-update to the given reference (commit, tag, or branch)
    Control who has access to trigger updates with the invoker permissions to the webhook
    """
    logger = utils.CloudLogger().log_request(request)
    try:
        # 1) Check token details to ensure user in whitelist
        token = utils.extract_token(request.headers, None)
        if token is None:
            return {
                'error': 'No credentials',
                'message': 'User did not pass identity in Authorization header'
            }, 400
        token_data = utils.get_token_info(token)
        if 'error' in token_data:
            return (
                {
                    'error': 'Invalid Token',
                    'message': token_data['error_description'] if 'error_description' in token_data else 'Google rejected the client token'
                },
                401
            )
        if token_data['email'] not in os.environ['INVOKERS'].split(','):
            return {
                'error': 'Not authorized',
                'message': 'User "{}" not present in function environment configuration'.format(token_data['email'])
            }, 403
        data = request.get_json()
        if 'ref' not in data:
            return {
                'error': 'Missing parameters',
                'message': 'Missing required parameter "ref"',
            }, 400

        update_payload = {
            'random': os.urandom(16).hex(),
            'tag': data['ref'],
            'url': 'https://github.com/broadinstitute/lapdog.git'
        }

        # 2) Get all resolved namespaces and update the iam policy for the signing key
        default_session = utils.generate_default_session()
        resolutions = [
            blob.download_as_string().decode()
            for page in utils._getblob_client(default_session.credentials).bucket('lapdog-resolutions').list_blobs(fields='items/name,nextPageToken').pages
            for blob in page
        ]

        policy = default_session.get(
            'https://cloudkms.googleapis.com/v1/projects/broad-cga-aarong-gtex/locations/global/keyRings/lapdog:getIamPolicy'
        )
        if policy.status_code != 200:
            return {
                'error': 'Unable to read IAM policy',
                'message':response.text
            }, 500
        policy = policy.json()
        for i, binding in enumerate(policy['bindings']):
            if binding['role'] == 'projects/broad-cga-aarong-gtex/roles/signingKeyVerifier':
                policy['bindings'][i]['members'] = [
                    'serviceAccount:lapdog-functions@{}.iam.gserviceaccount.com'.format(resolution)
                    for resolution in resolutions
                ]
        logger.log(
            "Updating project-wide IAM roles",
            bindings={
                'lapdog-functions@{}.iam.gserviceaccount.com'.format(resolution): 'projects/broad-cga-aarong-gtex/roles/signingKeyVerifier'
                for resolution in resolutions
            },
            severity='INFO'
        )
        response = default_session.post(
            'https://cloudkms.googleapis.com/v1/projects/broad-cga-aarong-gtex/locations/global/keyRings/lapdog:setIamPolicy',
            headers={'Content-Type': 'application/json'},
            json={
                "policy": policy,
                "updateMask": "bindings"
            }
        )
        if response.status_code != 200:
            return {
                'error': 'Unable to update IAM policy',
                'message': "Google rejected the policy update: (%d) : %s" % (response.status_code, response.text)
            }, 500

        # 3) Trigger update for all resolutions
        status = {
            'results': []
        }
        failed = 200
        logger.log(
            'Generating new signature',
            data=json.dumps(update_payload)
        )
        signature = utils._get_signature(json.dumps(update_payload).encode(), utils.UPDATE_KEY_PATH, default_session.credentials)
        max_version = int(utils.__API_VERSION__['update'][1:])
        for resolution in resolutions:
            updated = False
            for version in range(max_version, 0, -1):
                try:
                    update_url = 'https://us-central1-{project}.cloudfunctions.net/update-{version}'.format(
                        project=resolution,
                        version=
                    )
                    if default_session.options(update_url).status_code == 204:
                        logger.log(
                            "Triggering update",
                            project=resolution
                        )
                        response = default_session.post(
                            update_url,
                            headers={
                                'Content-Type': 'application/json',
                                'X-Lapdog-Signature': signature.hex()
                            },
                            json=update_payload
                        )
                        failed = max(failed, response.status_code)
                        status['results'].append({
                            'project': resolution,
                            'status': 'OK' if response.status_code == 200 else 'Failed',
                            'message': response.text,
                            'code': response.status_code
                        })
                        updated = True
                        break
                except:
                    logger.log_exception("Failed to update project", project=resolution)
                    failed = max(failed, 500)
                    status['results'].append({
                        'project': resolution,
                        'status': 'Error',
                        'message': traceback.format_exc(),
                        'code': 0
                    })
                    updated = True
                    break
            if not updated:
                status['results'].append({
                    'project': resolution,
                    'status': 'Error',
                    'message': "The target endpoint does not support and self-update endpoint versions",
                    'code': 0
                })

        return status, failed
    except:
        logger.log_exception()
        return {
            'error': 'Unknown error',
            'message': traceback.format_exc()
        }, 500

def trigger_update(ref):
    """
    Admins: Use to easily trigger the self-update webhook
    """
    return utils.generate_default_session(["https://www.googleapis.com/auth/cloud-platform"]).post(
        'https://us-central1-broad-cga-aarong-gtex.cloudfunctions.net/webhook-v1',
        headers={
            'Content-Type': 'application/json',
        },
        json={
            'ref': ref
        }
    )



import requests
import sys
from firecloud import api as fc
import subprocess
import io
import time
from functools import lru_cache
from flask import current_app
import random
import pandas as pd
import lapdog
import os
import json
import base64
import traceback
import select
from agutil import byteSize
from agutil.parallel import parallelize, parallelize2
from itertools import repeat
from glob import glob
import yaml
from .. import firecloud_status
from ..cache import cached, cache_fetch, cache_write, cache_init, cache_path
from ..adapters import NoSuchSubmission, Gateway, get_operation_status
from ..gateway import get_application_default_account, get_proxy_account
import re
import contextlib
import pickle
import tempfile
import html
from .. import getblob
from urllib.parse import unquote
from google.cloud import storage
from google.api_core.exceptions import Forbidden, BadRequest
from dalmatian.core import APIException


@parallelize2(1)
def get_womtool():
    data = cache_fetch('static', 'womtool', '39', 'jar', decode=False)
    if data is None:
        print("Downloading womtool for syntax highlighting")
        data = requests.get('https://github.com/broadinstitute/cromwell/releases/download/39/womtool-39.jar').content
        cache_write(data, 'static', 'womtool', '39', 'jar', decode=False)
    return cache_path('static')('womtool', '39', 'jar', dtype='data', ext='')

@contextlib.contextmanager
def log_controller(func):
    try:
        start = time.time()
        print("Started serving", func.__name__)
        yield
    finally:
        print("Finished serving", func.__name__, "in %0.1f seconds" % (time.time() - start))

def controller(func):
    def wrapper(*args, **kwargs):
        with log_controller(func):
            return func(*args, **kwargs)
    return wrapper

def readvar(obj, *args):
    current = obj
    for arg in args:
        if arg not in current:
            return None
        current = current[arg]
    return current

def get_workspace_object(namespace, name):
    ws = readvar(current_app.config, 'storage', 'cache', namespace, name, 'manager')
    if ws is None:
        ws = lapdog.WorkspaceManager("{}/{}".format(namespace, name), workspace_seed_url=None)
        if readvar(current_app.config, 'storage', 'cache') is None:
            current_app.config['storage']['cache'] = {}
        if readvar(current_app.config, 'storage', 'cache', namespace) is None:
            current_app.config['storage']['cache'][namespace] = {}
        if readvar(current_app.config, 'storage', 'cache', namespace, name) is None:
            current_app.config['storage']['cache'][namespace][name] = {}
        if readvar(current_app.config, 'storage', 'cache', namespace, name, 'manager') is None:
            current_app.config['storage']['cache'][namespace][name]['manager'] = ws
        workspaces = readvar(current_app.config, 'storage', 'cache', 'all_workspaces')
        if workspaces is None:
            current_app.config['storage']['cache']['all_workspaces'] = []
        current_app.config['storage']['cache']['all_workspaces'].append((namespace, name))
    return ws

@cached(60)
@controller
def get_alerts():
    return [
        json.loads(blob.download_as_string())
        for blob in storage.Client().bucket('lapdog-alerts').list_blobs()
    ]

@lru_cache()
@controller
def version():
    with open(os.path.join(os.path.dirname(os.path.dirname(__file__)), '__init__.py')) as reader:
        result = re.search(r'__version__ = \"(\d+\.\d+\.\d+.*?)\"', reader.read())
        if result:
            return result.group(1)
        return "Unknown"

@cached(120, 10)
def get_adapter(namespace, workspace, submission):
    return get_workspace_object(namespace, workspace).get_adapter(submission)

# @cached(60)
# def _get_lines(namespace, workspace, submission):
#     adapter = get_adapter(namespace, workspace, submission)
#     reader = adapter.read_cromwell()
#     return reader, []

def get_lines(namespace, workspace, submission):
    # from ..adapters import do_select
    # reader, lines = _get_lines(namespace, workspace, submission)
    # print("READER", reader)
    # while len(do_select(reader, 1)[0]):
    #     lines.append(reader.readline().decode().strip())
    #     print("READLINE", lines[-1])
    # return lines
    reader = get_adapter(namespace, workspace, submission).read_cromwell()
    lines = []
    from select import select
    try:
        if len(select([reader], [], [], 10)[0]):
            lines.append(reader.readline().decode().strip())
        while len(select([reader], [], [], 1)[0]):
            lines.append(reader.readline().decode().strip())
    except:
        lines.append([line.decode().strip() for line in reader.readlines()])
    return lines

@cached(30)
@controller
def status():
    try:
        return (
            {
                **{'failed': False},
                **firecloud_status()
            },
            200
        )
    except:
        print(traceback.format_exc())
    return {'failed': True, 'systems': {}}, 200

@cached(120)
@controller
def list_workspaces():
    try:
        return sorted([
            {
                'accessLevel': workspace['accessLevel'],
                'public': workspace['public'],
                'namespace': workspace['workspace']['namespace'],
                'name': workspace['workspace']['name'],
                'bucket': workspace['workspace']['bucketName'],
                'id': workspace['workspace']['workspaceId']
            } for workspace in fc.list_workspaces().json()
        ], key=lambda x:(x['namespace'], x['name'])), 200
    except:
        print(traceback.format_exc())
        all_workspaces = readvar(current_app.config, 'storage', 'cache', 'all_workspaces')
        if all_workspaces is not None:
            workspace_data = []
            for ns, ws, in all_workspaces:
                workspace = get_workspace_object(ns, ws).get_workspace_metadata()
                workspace_data.append({
                    'accessLevel': workspace['accessLevel'],
                    'public': False,
                    'namespace': ns,
                    'name': ws,
                    'bucket': workspace['workspace']['bucketName'],
                    'id': workspace['workspace']['workspaceId']
                })
            return sorted(workspace_data, key=lambda x:(x['namespace'], x['name'])), 200
    return {
        'failed': True,
        'reason': "The Firecloud api is currently offline, and there are no workspaces in the cache"
    }, 500

@cached(120)
@controller
def workspace(namespace, name):
    ws = get_workspace_object(namespace, name)
    data = {
        '__failures__': [],
    }
    try:
        data.update(ws.get_workspace_metadata())
        if 'workspace' in data and 'attributes' in data['workspace']:
            data['attributes'] = data['workspace']['attributes']
    except APIException:
        traceback.print_exc()
        # Fill in fallback data for UI to work
        data.update({
            '__failures__': ['metadata'],
            'workspace': {'bucketName': 'UNKNOWN'},
            'accessLevel': 'UNKNOWN',
            'owners': ['UNKNOWN'],
            'workspaceSubmissionStats': {'runningSubmissionsCount': 0}
        })
    try:
        data['entities'] = [
            {**v, **{'type':k}}
            for k,v in ws.get_entity_types().items()
        ]
    except APIException:
        traceback.print_exc()
        data['__failures__'].append('entities')
    try:
        data['configs'] = get_configs(namespace, name)
    except APIException:
        traceback.print_exc()
        data['__failures__'].append('configs')
    return data, 200


@cached(600)
@controller
def get_namespace_registered(namespace, name):
    ws = get_workspace_object(namespace, name)
    exists = ws.gateway.exists
    return {
        'exists': exists,
        'registered': exists and ws.gateway.registered,
        'compute_regions': ws.gateway.compute_regions if ws.gateway.exists else []
    }

@controller
def register(namespace, name):
    ws = get_workspace_object(namespace, name)
    if ws.gateway is not None and ws.gateway.exists and not ws.gateway.registered:
        ws.gateway.register(ws.workspace, ws.get_bucket_id())
        get_namespace_registered.cache_clear()
    return get_namespace_registered(namespace, name)



@cached(60)
@controller
def service_account():
    acct = get_proxy_account(get_application_default_account())
    if acct:
        return acct, 200
    return '<Not yet initialized. Register with any Lapdog Engine first>', 200

@cached(60)
@controller
def get_acl(namespace, name):
    failure = 'NONE'
    ws = get_workspace_object(namespace, name)
    try:
        acl = ws.acl
    except ValueError:
        traceback.print_exc()
        return (
            {
                'failed': True,
                'reason': 'permissions'
            },
            200
        )
    except NameError as e:
        traceback.print_exc()
        return (
            {
                'failed': True,
                'reason': 'firecloud',
                'error': e._request_text if hasattr(e, '_request_text') else "UNKNOWN"
            },
            200
        )
    except ValueError as e:
        traceback.print_exc()
        return (
            {
                'failed': True,
                'reason': 'firecloud',
                'error': e._request_text if hasattr(e, '_request_text') else "UNKNOWN"
            },
            200
        )
    return ({
        'accounts': [
            {
                'email': k,
                'access': v['accessLevel'],
                'share': v['canShare'],
                'compute': v['canCompute']
            }
            for k,v in acl.items()
        ],
        'failed': False,
        'reason': 'success'
    }, 200)

@cached(60)
@controller
def service_acl(namespace, name):
    data, code = get_acl(namespace, name)
    if code != 200 or (data['failed'] and data['reason'] != 'permissions'):
        return {
            'failed': True,
            'reason': data['reason']
        }, code
    perms = data['reason'] == 'permissions'
    account, code = service_account()
    if code != 200:
        print(account)
        return {
            'failed': True,
            'reason': 'gcloud',
        }, 200
    response = getattr(fc, '__get')('/api/proxyGroup/%s'%account)
    if response.status_code == 404:
        return {
            'failed': True,
            'reason': 'registration'
        }, 200
    elif response.status_code != 200:
        print(response.text)
        return {
            'failed': True,
            'reason': 'firecloud'
        }, 200
    ws, code = workspace(namespace, name)
    if code != 200:
        print(ws)
        return {
            'failed': True,
            'reason': 'acl-read'
        }, 200

    account_data = [acct for acct in data['accounts'] if acct['email'] == account] if 'accounts' in data else []

    return {
        'failed': perms,
        'reason': 'success' if not perms else 'permissions',
        'share': ws['canShare'],
        'service_account': len(account_data) and len({acct['access'] for acct in account_data} & {'WRITER', 'OWNER', 'ADMIN', 'PROJECT_OWNER'})
    }, 200

@controller
def set_acl(namespace, name):
    ws = get_workspace_object(namespace, name)
    account, code = service_account()
    if code != 200:
        print(account)
        return {
            'failed': True,
            'reason': 'gcloud',
        }, 200
    response = getattr(fc, '__get')('/api/proxyGroup/%s'%account)
    if response.status_code == 404:
        return {
            'failed': True,
            'reason': 'registration'
        }, 200
    elif response.status_code != 200:
        print(response.text)
        return {
            'failed': True,
            'reason': 'firecloud'
        }, 200
    acl, code = service_acl(namespace, name)
    if code != 200 or acl['failed']:
        print(acl, code)
        return {
            'failed': True,
            'reason': 'acl-read',
        }, 200
    if not acl['service_account']:
        if not acl['share']:
            return {
                'failed': True,
                'reason': 'permissions'
            }, 200
        try:
            data = ws.update_acl({
                account: 'WRITER'
            })
        except Exception as e:
            traceback.print_exc()
            return {
                'failed': True,
                'reason': 'firecloud',
                'error': e._request_text if hasattr(e, '_request_text') else 'UNKNOWN'
            }
        if 'usersUpdated' in data or 'invitesSent' in data or 'usersNotFound' in data:
            if 'usersNotFound' in data and account in {acct['email'] for acct in data['usersNotFound']}:
                return {
                    'failed': True,
                    'reason': 'registration'
                }, 200
            elif 'usersUpdated' not in data or account not in {acct['email'] for acct in data['usersUpdated']}:
                print(data)
                return {
                    'failed': True,
                    'reason': 'firecloud'
                }, 200
        else:
            print(data)
            return {
                'failed': True,
                'reason': 'firecloud'
            }, 200
    return {
        'failed': False,
        'reason': 'success',
        'share': True,
        'service_account': True
    }, 200

@cached(60)
def _get_entitites_df(namespace, name, etype):
    ws = get_workspace_object(namespace, name)
    return ws._get_entities_internal(etype)

@cached(10)
@controller
def get_entities(namespace, name, etype, start=0, end=None):
    # result = get_workspace_object(namespace, name).get_entity_types()
    # return {
    #     'failed': False,
    #     'reason': 'success',
    #     'entity_types': sorted([
    #         {
    #             'type': key,
    #             'attributes': val['attributeNames'],
    #             'n': val['count'],
    #             'id': val['idName'],
    #             # 'cache': get_cache(namespace, name, key)
    #         } for key, val in result.items()
    #     ], key=lambda x:x['type'])
    # }, 200
    df = _get_entitites_df(namespace, name, etype).reset_index()
    buffer = io.StringIO()
    if end is not None:
        df.iloc[start:end].to_json(buffer, 'records')
    else:
        df.to_json(buffer, 'records')
    return json.loads(buffer.getvalue())

@controller
def get_cache(namespace, name):
    ws = get_workspace_object(namespace, name)
    return {'state': ws.live, 'pending': len(ws.pending_operations)}, 200

@controller
def sync_cache(namespace, name):
    ws = get_workspace_object(namespace, name)
    if ws.live:
        ws.populate_cache()
        ws._webcache_ = True
        ws.go_offline()
    else:
        ws.sync()
    return get_cache(namespace, name)

@controller
def create_workspace(namespace, name, parent):
    ws = lapdog.WorkspaceManager("{}/{}".format(namespace,name), workspace_seed_url=None)
    parent = None if '/' not in parent else lapdog.WorkspaceManager(parent, workspace_seed_url=None)
    try:
        ws.create_workspace(parent)
        return {
            'failed': False,
            'reason': 'success'
        }, 200
    except APIException as e:
        try:
            reason = e.response.json()
            if 'message' in reason:
                reason = reason['message']
            return {
                'failed': True,
                'reason': reason
            }, 200
        except:
            pass
        return {
            'failed': True,
            'reason': traceback.print_exc()
        }, 200
    return {
        'failed': True,
        'reason': "Unknown"
    }, 200

@cached(60)
@controller
def get_configs(namespace, name):
    ws = get_workspace_object(namespace, name)
    get_womtool() # enque a download without waiting
    return ws.list_configs()

@cached(20)
@controller
def list_submissions(namespace, name, cache):
    from ..lapdog import timestamp_format
    ws = get_workspace_object(namespace, name)
    return sorted(
        (
            sub for sub in ws.list_submissions(lapdog_only=True, cached=cache)
            if 'identifier' in sub
        ),
        key=lambda s:(
            time.strptime(s['submissionDate'], timestamp_format)
            if s['submissionDate'] != 'TIME'
            else time.gmtime()
            ),
        reverse=True
    )

@controller
def preflight(namespace, name, config, entity, expression="", etype=""):
    ws = get_workspace_object(namespace, name)
    try:
        result = ws.preflight(
            config,
            entity,
            expression if expression != "" else None,
            etype if etype != "" else None
        )
        if result.result:
            return {
                'failed': False,
                'ok': result.result,
                'message': 'Okay',
                'workflows': len(result.workflow_entities),
                'invalid_inputs': '' # removed
            }, 200
        return {
            'failed': False,
            'ok': False,
            'message': result.reason,
            'workflows': 0,
            'invalid_inputs': 'None'
        }, 200
    except:
        print(sys.exc_info())
        return {
            'failed': True,
            'ok': False,
            'message': "Failure: "+traceback.format_exc(),
            'workflows': 0,
            'invalid_inputs': 'None'
        }, 200

@controller
def execute(namespace, name, config, entity, expression="", etype="", memory=3, batch=250, cache=True, private=False, region=None):
    ws = get_workspace_object(namespace, name)
    try:
        global_id, local_id, operation_id = ws.execute(
            config,
            entity,
            expression if expression != "" else None,
            etype if etype != "" else None,
            force=True,
            use_cache=cache,
            memory=memory,
            batch_limit=batch,
            private=private,
            region=region
        )
        return {
            'failed': False,
            'ok': True,
            'global_id': global_id,
            'local_id': local_id,
            'operation_id': operation_id
        }, 200
    except:
        traceback.print_exc()
        return {
            'failed': True,
            'ok': False,
            'message': traceback.format_exc(),
        }, 200

@controller
def decode_submission(submission_id):
    if submission_id.startswith('lapdog/'):
        ns, ws, sid = base64.b64decode(submission_id[7:].encode()).decode().split('/')
        return {
            'namespace': ns,
            'workspace': ws,
            'id': sid
        }, 200
    return "Not a valid submission id", 404

@controller
def get_submission(namespace, name, id):
    ws = get_workspace_object(namespace, name)
    try:
        adapter = get_adapter(namespace, name, id)
    except NoSuchSubmission:
        traceback.print_exc()
        return "No Such Submission", 404
    return {
        **ws.get_submission(id),
        **{
            'gs_path': os.path.join(
                'gs://'+ws.get_bucket_id(),
                'lapdog-executions',
                id
            ),
        }
    }, 200

@controller
def get_cost(namespace, name, id):
    ws = get_workspace_object(namespace, name)
    try:
        adapter = get_adapter(namespace, name, id)
    except NoSuchSubmission:
        traceback.print_exc()
        return "No Such Submission", 404
    return adapter.cost()

@controller
def abort_submission(namespace, name, id):
    adapter = get_adapter(namespace, name, id)
    adapter.abort()
    return [wf.long_id for wf in adapter.workflows.values() if wf.long_id is not None], 200

@controller
def upload_submission(namespace, name, id):
    ws = get_workspace_object(namespace, name)
    try:
        stats = ws.complete_execution(id)
    except FileNotFoundError:
        traceback.print_exc()
        return {
            'failed': True,
            'message': 'Job did not complete. It may have been aborted'
        }, 200
    except Exception as e:
        traceback.print_exc()
        return {
            'failed': True,
            'message': 'Exception ({}); Check your terminal for more info'.format(
                type(e)
            )
        }, 500
    else:
        return {
            'failed': not stats,
            'message': 'Failed'
        }, 200

@controller
def read_cromwell(namespace, name, id, offset=0):
    # _get_lines.cache_clear()
    lines = get_lines(namespace, name, id)
    return lines[offset:], 200
    # adapter = get_adapter(namespace, name, id)
    # reader = adapter.read_cromwell()
    # from ..adapters import do_select
    # lines = []
    # while len(do_select(reader, 1)[0]):
    #     lines.append(reader.readline().decode().strip())
    # return lines

@cached(10)
@controller
def get_workflows(namespace, name, id):
    adapter = get_adapter(namespace, name, id)
    adapter.update()
    return [
        {
            'entity': wf['workflowEntity'],
            'id': adapter.workflow_mapping[wf['workflowOutputKey']],
            'status': (
                adapter.workflows[adapter.workflow_mapping[wf['workflowOutputKey']][:8]].status
                if adapter.workflow_mapping[wf['workflowOutputKey']][:8] in adapter.workflows
                else 'Starting'
            )
        }
        for wf in adapter.raw_workflows
        if wf['workflowOutputKey'] in adapter.workflow_mapping
    ], 200

# @cached(10)
@cached(10)
@controller
def get_workflow(namespace, name, id, workflow_id):
    adapter = get_adapter(namespace, name, id)
    adapter.update(5)
    # print(adapter.workflows)
    if workflow_id[:8] in adapter.workflows:
        # Return data from workflow
        wf = adapter.workflows[workflow_id[:8]]
        entity_map = {w['workflowOutputKey']:w['workflowEntity'] for w in adapter.raw_workflows}
        reverse_map = {v:entity_map[k] for k,v in adapter.workflow_mapping.items()}
        workflow_inputs = None
        try:
            workflow_inputs = wf.inputs
        except:
            traceback.print_exc()
        return {
            'id': wf.long_id,
            'short_id': wf.id,
            'key': wf.key,
            'running': wf.started,
            'n_calls': len(wf.calls),
            'calls': [
                {
                    'status': call.status,
                    'message': call.last_message,
                    'attempt': call.attempt,
                    'operation': call.operation,
                    'task': call.task,
                    'gs_path': call.path,
                    'code': call.return_code,
                    'idx': i,
                    'runtime': call.runtime
                }
                for i, call in enumerate(wf.calls)
            ],
            'gs_path': wf.path,
            'status': wf.status,
            'failure': wf.failure,
            'entity': (
                reverse_map[workflow_id]
                if workflow_id in reverse_map
                else None
            ),
            'inputs': workflow_inputs
        }, 200
    else:
        return {
            'id': None,
            'short_id': None,
            'key': None,
            'running': False,
            'n_calls': 0,
            'calls': [],
            'gs_path': None,
            'status': 'Pending',
            'failure': None,
            'entity': None,
            'inputs': None
        }, 200

# @cached(30)
@controller
def read_logs(namespace, name, id, workflow_id, log, call):
    log_text = cache_fetch('workflow', id, workflow_id, dtype=str(call)+'.', ext=log+'.log')
    if log_text is not None:
        return log_text, 200
    adapter = get_adapter(namespace, name, id)
    adapter.update()
    if workflow_id[:8] in adapter.workflows:
        workflow = adapter.workflows[workflow_id[:8]]
        if call < len(workflow.calls):
            call = workflow.calls[call]
            try:
                text = call.read_log(log)
                return text, 200
            except FileNotFoundError:
                return "Not Found", 404
    return "Error", 500

@cached(10)
@controller
def operation_status(operation_id):
    return get_operation_status(operation_id, False, 'yaml'), 200

@controller
def cache_size():
    total = 0
    for path, _, files in os.walk(cache_init()):
        for f in files:
            total += os.path.getsize(os.path.join(path, f))
    return byteSize(total), 200

@cached(120)
@controller
def quotas(namespace):
    try:
        return Gateway(namespace).quota_usage
    except NameError:
        traceback.print_exc()
        print("No resolution found")
        return {
            'raw': [],
            'alerts': []
        }


@cached(60)
@controller
def get_config(namespace, name, config_namespace, config_name):
    ws = get_workspace_object(namespace, name)
    try:
         config = ws.get_config("{}/{}".format(config_namespace, config_name))
    except:
        ws.sync()
        return (
            {
                'config': None,
                'wdl': None
            }, 200)
    try:
        data = getattr(fc, '__post')(
            '/api/inputsOutputs',
            data=json.dumps(config['methodRepoMethod'])
        ).json()
        io_types = {
            'inputs': {
                param['name']:{
                    'type': param['inputType'],
                    'required': not param['optional']
                }
                for param in data['inputs']
            },
            'outputs': {
                param['name']:param['outputType']
                for param in data['outputs']
            }
        }
    except KeyError:
        traceback.print_exc()
        print("The above error is likely caused by a missing WDL", file=sys.stderr)
        io_types = None
    except:
        traceback.print_exc()
        io_types = None
    try:
        wdl_text = ws.get_wdl(
            config['methodRepoMethod']['methodPath']
            if 'sourceRepo' in config['methodRepoMethod'] and config['methodRepoMethod']['sourceRepo'] == 'dockstore'
            else "{}/{}".format(config['methodRepoMethod']['methodNamespace'], config['methodRepoMethod']['methodName'])
        )
        try:
            with tempfile.NamedTemporaryFile('w', suffix='wdl') as tmpwdl:
                tmpwdl.write(wdl_text)
                tmpwdl.flush()
                result = subprocess.run(
                    'java -jar {} highlight {} --highlight-mode html'.format(get_womtool()(), tmpwdl.name),
                    shell=True,
                    executable='/bin/bash',
                    stdout=subprocess.PIPE
                )
                if result.returncode == 0:
                    raw_text = result.stdout.decode()
                    wdl_text = ''
                    pattern = re.compile(r'</?span.*?>')
                    while pattern.search(raw_text):
                        match = pattern.search(raw_text)
                        wdl_text += (
                            html.escape(raw_text[:match.start()])
                            + match.group()
                        )
                        raw_text = raw_text[match.end():]
                    wdl_text += html.escape(raw_text)
        except:
            traceback.print_exc()
        return {
            'config':config,
            'io': io_types,
            'wdl': wdl_text
        }, 200
    except:
        traceback.print_exc()
        return {
            'config':config,
            'wdl': None
        }, 200
    return "FAILED", 500

@controller
def update_config(namespace, name, config):
    ws = get_workspace_object(namespace, name)
    ws.update_config(config)
    get_config.cache_clear()
    get_configs.cache_clear()
    return ("OK", 200)

@controller
def delete_config(namespace, name, config_namespace, config_name):
    ws = get_workspace_object(namespace, name)
    ws.delete_config("{}/{}".format(config_namespace, config_name))
    get_config.cache_clear()
    get_configs.cache_clear()
    return ("OK", 200)

@controller
def upload_config(namespace, name, config_filepath, method_filepath=None):
    try:
        config = json.load(config_filepath)
    except:
        traceback.print_exc()
        return {
            'failed': True,
            'reason': "Unable to parse config json"
        }, 200
    try:
        ws = get_workspace_object(namespace, name)
        live = ws.live
        result = ws.update_config(
            config,
            method_filepath, # Either it's a file-object or None
        )
        get_config.cache_clear()
        get_configs.cache_clear()
        return {
            'failed': not result,
            'reason': 'Success' if result else (
                'API rejected update' if live else 'Workspace Offline. WDL cannot be synced to FireCloud'
            )
        }
    except KeyError:
        traceback.print_exc()
        return {
            'failed': True,
            'reason': "Configuration missing required keys"
        }, 200
    except:
        traceback.print_exc()
        return {
            'failed': True,
            'reason': "Unhandled exception. Check Lapdog terminal output for details"
        }, 200

@controller
def rerun_submission(namespace, name, id):
    ws = get_workspace_object(namespace, name)
    return {
        **{
            'name': None,
            'type': None,
            'expression': None
        },
        **ws.build_retry_set(id)
    }

@cached(120)
@controller
def config_autocomplete(namespace, name, config_namespace, config_name):
    ws = get_workspace_object(namespace, name)
    entityTypeMetadata = ws.get_entity_types()[ws.get_config("{}/{}".format(config_namespace, config_name))['rootEntityType']]
    return (
        [
            'workspace.'+attribute for attribute in ws.get_attributes()
        ] + [
        'this.'+attribute for attribute in entityTypeMetadata['attributeNames']
        ] + ['this.name', 'this.'+entityTypeMetadata['idName']]
    ), 200

@controller
def autocomplete(namespace, name, entity):
    """
    Entrypoint for autocomplete system
    Uses heavy caching to accelerate results
    1) Checks for previous autocomplete searches in this workspace which could serve as a starting point
    2) For any searches which are a substring of the current search, fetch those results and then filter using the current search
    3) Otherwise, start from scratch (fetching new entities again) and filter using the current search
    """
    if len(entity) <= 1:
        return []
    for search in [*get_autocomplete_cache(namespace, name)]:
        if search in entity:
            return [
                eid for eid in get_autocomplete_entries(namespace, name, search)
                if entity in eid
            ]
    return get_autocomplete_entries(namespace, name, entity)

@cached(60)
def get_autocomplete_cache(namespace, name):
    return set()

@cached(60, cache_size=64)
def get_autocomplete_entries(namespace, name, entity):
    ws = get_workspace_object(namespace, name)
    get_autocomplete_cache(namespace, name).add(entity)
    return [
        eid
        for etype in ws.get_entity_types()
        for eid in ws._get_entities_internal(etype).index
        if entity in eid
    ]

@controller
def seed_cache(namespace, name):
    ws = get_workspace_object(namespace, name)
    return {
        key:base64.b64encode(pickle.dumps(value)).decode()
        for key, value in ws.cache.items()
    }

__USER_PROJECT = None

def get_user_project(project=None):
    global __USER_PROJECT
    if project is not None:
        __USER_PROJECT = project
    else:
        if __USER_PROJECT is None:
            proj = subprocess.run(
                'gcloud config get-value billing/quota_project',
                shell=True,
                stdout=subprocess.PIPE
            ).stdout.decode().strip()
            if proj == 'CURRENT_PROJECT' or proj == '':
                proj = subprocess.run(
                    'gcloud config get-value project',
                    shell=True,
                    stdout=subprocess.PIPE
                ).stdout.decode().strip()
                __USER_PROJECT = (
                    proj if proj != '' else None
                )

    return __USER_PROJECT


@cached(15)
@controller
def preview_blob(path, project=None):
    return _preview_blob_internal(
        unquote(path.replace('~', '%')),
        project
    )

def _preview_blob_internal(path, project):
    if not path.startswith('gs://'):
        path = 'gs://'+path
    blob = getblob(path, user_project=project)
    try:
        try:
            exists = blob.exists()
        except ValueError:
            exists = False
        blob.bucket.reload()
        if exists: # file
            blob.reload()
            return (
                {
                    'requesterPays': blob.bucket.requester_pays,
                    'exists': exists,
                    'size': byteSize(blob.size),
                    'url': None, # Temporary. Need a way to generate signed urls
                    'preview': blob.download_as_string(end=1024).decode('ascii', 'replace'),
                    'visitUrl': "https://console.cloud.google.com/storage/browser/{}/{}".format(
                        blob.bucket.name,
                        os.path.dirname(blob.name)
                    ),
                    'children': [],
                    'bucket': blob.bucket.name
                },
                200
            )
        elif len([*blob.bucket.list_blobs(prefix=blob.name, max_results=1)]):
            # is a directory
            return (
                {
                    'requesterPays': blob.bucket.requester_pays,
                    'exists': False,
                    'size': None,
                    'preview': None,
                    'url': None,
                    'visitUrl': "https://console.cloud.google.com/storage/browser/{}/{}".format(
                        blob.bucket.name,
                        blob.name
                    ),
                    'children': sorted({
                        os.path.relpath(child.name, blob.name).split('/')[0]
                        for page in blob.bucket.list_blobs(prefix=blob.name, fields='items/name,nextPageToken').pages
                        for child in page
                    }),
                    'bucket': blob.bucket.name
                }
            )
    except BadRequest as e:
        if 'requester pays' in e.message.lower():
            project = get_user_project(project)
            if project is not None:
                return _preview_blob_internal(path, project)
            return (
                {
                    'requesterPays': True,
                    'exists': False,
                    'size': None,
                    'preview': None,
                    'url': None,
                    'visitUrl': "https://console.cloud.google.com/storage/browser/{}/{}".format(
                        blob.bucket.name,
                        os.path.dirname(blob.name)
                    ),
                    'children': [],
                    'bucket': blob.bucket.name
                },
                200
            )
        traceback.print_exc()
        return (
            {
                "error": "Unable to query path",
                "message": traceback.format_exc()
            },
            400
        )
    except Forbidden:
        traceback.print_exc()
    except:
        traceback.print_exc()
        return (
            {
                "error": "Unable to query path",
                "message": traceback.format_exc()
            },
            500
        )
    return (
        {
            'requesterPays': False,
            'exists': False,
            'size': None,
            'preview': None,
            'url': None,
            'visitUrl': "https://console.cloud.google.com/storage/browser/{}/{}".format(
                blob.bucket.name,
                os.path.dirname(blob.name)
            ),
            'children': [],
            'bucket': blob.bucket.name
        },
        200
    )



# Submission Gateway
# Use as a modular component for adapter to interact with submission
# If submission JSON does not contain version field or version is 1, gcloud compute instances ssh
# If submission JSON contains version and version is 2, ssh -i {lapdog ssh token} {instance ip}
# If SSH fails for any reason, save the exception and try reading the log
# If the log is not found, then you're SOL

from functools import lru_cache
from firecloud import api as fc
import requests
import subprocess
from hashlib import md5, sha512
from .cache import cached, cache_fetch, cache_write
import time
import warnings
import contextlib
import crayons
import os
import io
import json
from threading import RLock
from .cloud import RESOLUTION_URL
from .cloud.utils import get_token_info, ld_project_for_namespace, ld_meta_bucket_for_project, proxy_group_for_user, generate_default_session, update_iam_policy, enabled_regions, __API_VERSION__, GCP_ZONES
from urllib.parse import quote
import google.api_core.exceptions
import sys
import tempfile
import base64
from agutil import cmd as run_cmd
from dalmatian import WorkspaceManager, getblob
import traceback
import re
from uuid import uuid4

CORE_PERMISSIONS = [
    "cloudkms.cryptoKeyVersions.useToSign",
    "cloudkms.cryptoKeyVersions.viewPublicKey",
    "cloudkms.cryptoKeyVersions.get",
    "cloudkms.cryptoKeyVersions.list",
    "resourcemanager.projects.get",
    "genomics.datasets.create",
    "genomics.datasets.delete",
    "genomics.datasets.get",
    "genomics.datasets.list",
    "genomics.datasets.update",
    "genomics.operations.cancel",
    "genomics.operations.create",
    "genomics.operations.get",
    "genomics.operations.list"
]

FUNCTIONS_PERMISSIONS = [
    "cloudkms.cryptoKeyVersions.viewPublicKey",
    "cloudkms.cryptoKeyVersions.get",
    "cloudkms.cryptoKeyVersions.list",
    "iam.serviceAccountKeys.create",
    "iam.serviceAccountKeys.delete",
    "iam.serviceAccountKeys.get",
    "iam.serviceAccountKeys.list",
    "iam.serviceAccounts.create",
    "iam.serviceAccounts.delete",
    "iam.serviceAccounts.get",
    "iam.serviceAccounts.getIamPolicy",
    "iam.serviceAccounts.list",
    "iam.serviceAccounts.setIamPolicy",
    "iam.serviceAccounts.update",
    "resourcemanager.projects.get",
    "genomics.operations.create",
    "genomics.operations.get",
    "genomics.operations.list",
    "resourcemanager.projects.getIamPolicy",
    "resourcemanager.projects.setIamPolicy",
    "compute.projects.get",
    "compute.regions.get",
    "compute.subnetworks.setPrivateIpGoogleAccess",
    "logging.logEntries.create"
]

ADMIN_PERMISSIONS = [
    "cloudfunctions.functions.call",
    "cloudfunctions.functions.create",
    "cloudfunctions.functions.get",
    "cloudfunctions.functions.list",
    "cloudfunctions.functions.sourceCodeGet",
    "cloudfunctions.functions.sourceCodeSet",
    "cloudfunctions.functions.update",
    "cloudfunctions.operations.get",
    "cloudfunctions.operations.list",
    "compute.networks.get",
    "compute.networks.list",
    "compute.networks.use",
    "compute.subnetworks.get",
    "compute.subnetworks.list",
    "compute.subnetworks.setPrivateIpGoogleAccess",
    "compute.subnetworks.update",
    "iam.roles.create",
    "iam.roles.get",
    "iam.roles.list",
    "iam.roles.update",
    "iam.serviceAccounts.get",
    "iam.serviceAccounts.list",
    "iam.serviceAccounts.create",
    'iam.serviceAccounts.setIamPolicy',
    'iam.serviceAccounts.getIamPolicy',
    "resourcemanager.projects.setIamPolicy",
    "resourcemanager.projects.getIamPolicy",
    "storage.objects.setIamPolicy",
    "storage.objects.update",
    "storage.buckets.get",
    "storage.objects.create",
    "storage.objects.delete",
    "storage.objects.get",
    "storage.objects.getIamPolicy",
    "storage.objects.list",
    "serviceusage.quotas.get",
    "serviceusage.quotas.update"
]

PET_PERMISSIONS = [
    "cloudkms.cryptoKeyVersions.viewPublicKey",
    "resourcemanager.projects.get",
    "genomics.operations.cancel",
    "genomics.operations.create",
    "genomics.operations.get",
    "genomics.operations.list",
    "serviceusage.services.use", # for requester pays
    "logging.logEntries.create"
]

_ACL_LOCK = RLock()

creation_success_pattern = re.compile(r'Workspace (.+)/(.+) successfully')

id_rsa = os.path.join(
    os.path.expanduser('~'),
    '.ssh',
    'id_rsa'
)

credentials_file = os.path.join(
    os.path.expanduser('~'),
    '.config',
    'gcloud',
    'application_default_credentials.json'
)

@cached(60, 1)
def get_gcloud_account():
    """
    Gets the currently logged in gcloud account.
    This checks main credentials (gcloud auth login)
    Not application default credentials used by lapdog (gcloud auth application-default login)
    60 second cache (because user can change sign-in)
    """
    return subprocess.run(
        'gcloud config get-value account',
        shell=True,
        stdout=subprocess.PIPE
    ).stdout.decode().strip()

def get_proxy_account(user_acct=None):
    if user_acct is None:
        user_acct = get_application_default_account()
    email = cache_fetch('proxy-acct', user_acct)
    if email is not None:
        return email
    acct = proxy_group_for_user(user_acct)
    response = fc.__get('/api/groups/{}'.format(acct))
    if response.status_code == 200:
        email = response.json()['groupEmail']
        if email == acct + '@firecloud.org':
            cache_write(email, 'proxy-acct', user_acct)
            return email

@lru_cache()
def get_application_default_account():
    """
    Gets the currently logged in gcloud application-default account.
    This checks application default credentials used by lapdog (gcloud auth application-default login)
    Not main credentials (gcloud auth login)
    LRU-cache because response cannot change after session is generated
    """
    return get_token_info(get_user_session())['email']

def get_access_token(account=None):
    """
    Gets an access token to authenticate the Lapdog Gateway as the given account.
    The `account` argument is deprecated and should be left as None.
    The most recent access token for a given account is stored in the offline cache.
    A new token is generated when the cached token expired (~1hr after generation).
    """
    with _ACL_LOCK:
        warnings.warn(
            "lapdog.gateway.get_access_token is deprecated for security reasons and will be removed soon."
            " Please transition to use lapdog.gateway.get_user_session() (or lapdog.cloud.utils.generate_default_session() if"
            " you need control over OAuth scopes)",
            DeprecationWarning,
            stacklevel=2
        )
        if account is None:
            account = get_application_default_account()
        token = cache_fetch('token', 'access-token', md5(account.encode()).hexdigest(), google='credentials')
        if token and get_token_expired(token):
            token = None
        if token is None:
            if not os.path.isfile(credentials_file):
                raise FileNotFoundError("Application Default Credentials not found. Please run `gcloud auth application-default login`")
            with open(credentials_file) as r:
                credentials = json.load(r)
            response = requests.post(
                'https://www.googleapis.com/oauth2/v4/token',
                data={
                    'client_id': credentials['client_id'],
                    'client_secret': credentials['client_secret'],
                    'refresh_token': credentials['refresh_token'],
                    'grant_type': 'refresh_token'
                }
            )
            if response.status_code == 200:
                data = response.json()
                token = data['access_token']
                expiry = int(time.time() + int(data['expires_in']))
                cache_write(token, 'token', 'access-token', md5(account.encode()).hexdigest(), google='credentials')
                cache_write(str(expiry), 'token', 'expiry', token=md5(token.encode()).hexdigest())
            else:
                raise ValueError("Unable to refresh access token (%d) : %s" % (response.status_code, response.text))
        return token

def get_token_expired(token):
    """
    Returns whether or not the current token has expired
    """
    expiry = cache_fetch('token', 'expiry', token=md5(token.encode()).hexdigest())
    if expiry is None:
        try:
            data = get_token_info(token)
            if data is None:
                return True
            expiry = int(time.time() + int(data['expires_in']))
            cache_write(str(expiry), 'token', 'expiry', token=md5(token.encode()).hexdigest())
        except:
            return True
    return int(expiry) < time.time()

@contextlib.contextmanager
def capture(display=True):
    """
    Context manager to redirect stdout and err
    """
    try:
        stdout_buff = io.StringIO()
        stderr_buff = io.StringIO()
        with contextlib.redirect_stdout(stdout_buff):
            with contextlib.redirect_stderr(stderr_buff):
                yield (stdout_buff, stderr_buff)
    finally:
        stdout_buff.seek(0,0)
        stderr_buff.seek(0,0)
        if display:
            print(stderr_buff.read(), end='', file=sys.stderr)
            stderr_buff.seek(0,0)
            print(stdout_buff.read(), end='')
            stdout_buff.seek(0,0)

def _generate_core_key_internal(ld_project, worker_account, session=None):
    """
    Issues a new core service account access key for a given lapdog engine project.
    This cannot be used unless you are an administrator.
    There is no reason to use this function unless:
    1) The current access key has expired (~10yr after generation)
    2) Someone has gained unauthorized access to the current key

    In the second case, you should log into the cloud console and remove any current
    keys for the core service account before calling this function. You should also
    check the acl bindings to ensure that no unathorized users have access to the metadata bucket
    either directly on object in the bucket or inheriting through project-level permissions
    You should also add a new key version of the lapdog-sign key in google KMS and
    remove all existing versions
    """
    if session is None:
        session = get_user_session()
    warnings.warn("Generating new root authentication key for project")
    response = session.post(
        "https://iam.googleapis.com/v1/projects/{project}/serviceAccounts/{account}/keys".format(
            project=quote(ld_project, safe=''),
            account=quote('lapdog-worker@{}.iam.gserviceaccount.com'.format(ld_project), safe='')
        )
    )
    if response.status_code == 200:
        blob = getblob(
            'gs://{bucket}/auth_key.json'.format(
                bucket=ld_meta_bucket_for_project(ld_project)
            )
        )
        blob.upload_from_string(
            base64.b64decode(response.json()['privateKeyData'].encode())
        )
        acl = blob.acl
        for entity in acl.get_entities():
            if entity.type == 'project':
                if entity.identifier.startswith('editors-'):
                    entity.revoke_owner()
                elif entity.identifier.startswith('viewers-'):
                    entity.revoke_read()
        acl.user(worker_account).grant_read()
        acl.save()
    else:
        print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
        raise ValueError("Could not generate core key")

def resolve_project_for_namespace(namespace):
    resolution = cache_fetch('namespace', 'resolution', namespace=namespace)
    if resolution is None:
        blob = getblob(
            'gs://lapdog-resolutions/' + sha512(namespace.encode()).hexdigest()
        )
        if not blob.exists():
            raise NameError("No resolution for "+namespace)
        resolution = blob.download_as_string().decode()
        try:
            # Before caching the resolution, check the inverse
            # The resolved project should report that it manages the provided namespace
            # It would be unusual if it did not, perhaps a resolution was manually changed
            local_blob = getblob(
                'gs://{}/resolution'.format(ld_meta_bucket_for_project(resolution))
            )
            if local_blob.exists() and local_blob.download_as_string().decode() == namespace:
                cache_write(resolution, 'namespace', 'resolution', namespace=namespace)
        except:
            traceback.print_exc()
    return resolution

_USER_SESSION_INTERNAL = None

def get_user_session():
    global _USER_SESSION_INTERNAL
    with _ACL_LOCK:
        if _USER_SESSION_INTERNAL is None:
            _USER_SESSION_INTERNAL = generate_default_session(scopes=[
                'profile',
                'email',
                'openid',
                'https://www.googleapis.com/auth/devstorage.read_write',
                'https://www.googleapis.com/auth/userinfo.profile',
                'https://www.googleapis.com/auth/userinfo.email'
            ])
            try:
                from hound.client import _getblob_bucket
                for blob in _getblob_bucket(None, 'lapdog-alerts', None).list_blobs():
                    content = json.loads(blob.download_as_string().decode())
                    if content['type'] == 'critical':
                        text = content['text'] if 'text' in content else content['content']
                        print(crayons.red("Critical Alert:"), text)
            except:
                traceback.print_exc()
                warnings.warn("Unable to check Lapdog global alerts during startup")
        return _USER_SESSION_INTERNAL


class Gateway(object):
    """
    Acts as an interface between local lapdog and any resources behind the project's API
    """
    def __init__(self, namespace):
        self.namespace = namespace
        try:
            self.project = resolve_project_for_namespace(namespace)
        except:
            self.project = None
            traceback.print_exc()
        try:
            if not self.exists:
                warnings.warn("Gateway does not exist. You will not be able to execute jobs")
            elif not self.registered:
                warnings.warn(
                    "Gateway for namespace {} not registered. Please run Gateway.register()".format(
                        self.namespace
                    )
                )
        except ValueError:
            traceback.print_exc()
            print("Gateway not up to date", file=sys.stderr)

    @classmethod
    def initialize_lapdog_for_project(cls, billing_id, project_id, custom_lapdog_project=None):
        """
        Initializes the lapdog execution API on the given firecloud project.
        Charges for operating the Lapdog API and for executing jobs will be billed
        to the provided billing id.
        The current gcloud account (when this function is executed) will be the Owner
        of the service and the only user capable of using it.
        Specify custom_lapdog_project to override default namespace->project association
        """
        if custom_lapdog_project is None:
            custom_lapdog_project = ld_project_for_namespace(project_id)
        print("Checking prerequisites")
        cmd = (
            'gcloud --version'
        )
        print(cmd)
        result = run_cmd(cmd)
        gcloud_version = {}
        for line in result.buffer.decode().split('\n'):
            if len(line.strip()):
                words = line.strip().split()
                gcloud_version[' '.join(words[:-1])] = int(words[-1].split('.')[0])
        if gcloud_version['Google Cloud SDK'] < 241:
            print("Update gcloud SDK? Lapdog requires >= 241.0.0")
            choice = input("Y/N : ")
            if not choice.strip().lower().startswith('y'):
                raise ValueError("Lapdog requires gcloud version >= 241.0.0; Please run 'gcloud components update'")
            print('gcloud components update')
            run_cmd('gcloud components update')
        if not ('alpha' in gcloud_version and 'beta' in gcloud_version):
            print("Install additional components? Lapdog requires gcloud alpha and beta")
            choice = input("Y/N : ")
            if not choice.strip().lower().startswith('y'):
                raise ValueError("Lapdog requires gcloud alpha and beta; Please run 'gcloud components install alpha beta'")
            print('gcloud components install alpha beta')
            run_cmd('gcloud components install alpha beta')

        print("Testing permissions")
        test_url = "https://cloudbilling.googleapis.com/v1/billingAccounts/{billing_account}:testIamPermissions".format(
            billing_account=billing_id
        )
        print("POST", test_url)
        user_session = generate_default_session()
        response = user_session.post(
            test_url,
            headers={"Content-Type": "application/json"},
            json={
                "permissions": [
                    "billing.accounts.get",
                    "billing.accounts.getIamPolicy",
                    "billing.resourceAssociations.create"
                ]
            }
        )
        if response.status_code != 200:
            raise ValueError("Unexpected response from Google API: ({}): {}".format(response.status_code, response.text))
        permissions = response.json()
        if 'permissions' not in permissions:
            raise ValueError("Unexpected response from Google API: %s. You may not have appropriate permissions for the billing account" % response.text)
        permissions = permissions['permissions']
        if "billing.accounts.get" not in permissions or "billing.accounts.getIamPolicy" not in permissions or "billing.resourceAssociations.create" not in permissions:
            raise ValueError("Insufficient permissions to use this billing account")

        print("GET https://api.firecloud.org/api/profile/billing")
        while True:
            response = user_session.get(
                'https://api.firecloud.org/api/profile/billing'
            )

            if response.status_code == 200:
                break
            print(response.status_code, response.text, file=sys.stderr)
            if response.status_code == 404:
                return (
                    {
                        'error': "User not found",
                        'message': "You are not registered yet with firecloud"
                    },
                    404
                )
            time.sleep(5)


        projects = {proj['projectName']:proj for proj in response.json()}
        if project_id not in projects:
            raise NameError("The provided namespace '%s' could not be found" % project_id)

        if projects[project_id]['role'] not in {'Owner', 'Admin', 'Administrator'}:
            raise ValueError("You lack Owner/Admin privilages on the provided namespace")

        cmd = (
            'gcloud projects create {project_id}'.format(
                project_id=custom_lapdog_project
            )
        )
        print("Creating project")
        print(cmd)
        result = run_cmd(cmd)
        if result.returncode != 0:
            if b'already in use' not in result.buffer:
                raise ValueError("Unable to create project")
            else:
                print("This project is already in use")
                print("Do you wish to continue initialization with the given project?")
                choice = input("Y/N : ")
                if not choice.strip().lower().startswith('y'):
                    print("Aborted")
                    return
        cmd = 'gcloud --project {project} services enable cloudbilling.googleapis.com'.format(
            project=custom_lapdog_project,
        )
        print(cmd)
        subprocess.check_call(cmd, shell=True)
        cmd = 'gcloud --project {project} services enable serviceusage.googleapis.com'.format(
            project=custom_lapdog_project,
        )
        print(cmd)
        subprocess.check_call(cmd, shell=True)
        cmd = (
            'gcloud beta billing projects link {project_id} --billing-account '
            '{billing_id}'.format(
                project_id=custom_lapdog_project,
                billing_id=billing_id
            )
        )
        print("Enabling billing")
        print(cmd)
        subprocess.check_call(cmd, shell=True)
        print("Saving Namespace Resolution")
        response = user_session.post(
            RESOLUTION_URL,
            headers={"Content-Type": "application/json"},
            json={
                'namespace': project_id,
                'project': custom_lapdog_project
            }
        )
        if response.status_code == 409:
            proj = resolve_project_for_namespace(project_id)
            if proj != custom_lapdog_project:
                raise NameError("A resolution is already in place for this namespace. Please contact GitHub @agraubert")
        elif response.status_code != 200:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Error when generating resolution")
        print("Enabling servies...")
        services = [
            'cloudapis.googleapis.com',
            'clouddebugger.googleapis.com',
            'cloudfunctions.googleapis.com',
            'cloudkms.googleapis.com',
            'cloudresourcemanager.googleapis.com',
            'cloudtrace.googleapis.com',
            'compute.googleapis.com',
            'deploymentmanager.googleapis.com',
            'genomics.googleapis.com',
            'iam.googleapis.com',
            'iamcredentials.googleapis.com',
            'logging.googleapis.com',
            'servicemanagement.googleapis.com',
            'storage-component.googleapis.com',
            'storage-api.googleapis.com'
        ]
        for service in services:
            cmd = 'gcloud --project {project} services enable {service}'.format(
                project=custom_lapdog_project,
                service=service
            )
            print(cmd)
            subprocess.check_call(cmd, shell=True)
        print("Creating Signing Key")
        cmd = (
            'gcloud --project {project} kms keyrings create lapdog --location us'.format(
                project=custom_lapdog_project
            )
        )
        print(cmd)
        result = run_cmd(cmd)
        if result.returncode != 0 and b'ALREADY_EXISTS' not in result.buffer:
            raise ValueError("Unable to create Keyring")
        cmd = (
            'gcloud --project {project} alpha kms keys create lapdog-sign --location us --keyring'
            ' lapdog --purpose asymmetric-signing --default-algorithm '
            'rsa-sign-pss-3072-sha256 --protection-level software'.format(
                project=custom_lapdog_project
            )
        )
        print(cmd)
        result = run_cmd(cmd)
        if result.returncode != 0 and b'ALREADY_EXISTS' not in result.buffer:
            raise ValueError("Unable to create signing key")

        print("Creating Lapdog Roles")
        roles_url = "https://iam.googleapis.com/v1/projects/{project}/roles".format(
            project=custom_lapdog_project
        )
        print("POST", roles_url)
        response = user_session.post(
            roles_url,
            headers={
                'Content-Type': 'application/json'
            },
            json={
                "roleId": "Pet_account",
                "role": {
                    "title": "Pet_account",
                    "includedPermissions": PET_PERMISSIONS,
                    "stage": "GA"
                }
            }
        )
        if response.status_code != 200 and 'ALREADY_EXISTS' not in response.text:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Invalid response from Google API")
        print("POST", roles_url)
        response = user_session.post(
            roles_url,
            headers={
                'Content-Type': 'application/json'
            },
            json={
                "roleId": "Core_account",
                "role": {
                    "title": "Core_account",
                    "includedPermissions": CORE_PERMISSIONS,
                    "stage": "GA"
                }
            }
        )
        if response.status_code != 200 and 'ALREADY_EXISTS' not in response.text:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Invalid response from Google API")
        print("POST", roles_url)
        response = user_session.post(
            roles_url,
            headers={
                'Content-Type': 'application/json'
            },
            json={
                "roleId": "Functions_account",
                "role": {
                    "title": "Functions_account",
                    "includedPermissions": FUNCTIONS_PERMISSIONS,
                    "stage": "GA"
                }
            }
        )
        if response.status_code != 200 and 'ALREADY_EXISTS' not in response.text:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Invalid response from Google API")
        print("POST", roles_url)
        response = user_session.post(
            roles_url,
            headers={
                'Content-Type': 'application/json'
            },
            json={
                "roleId": "Engine_Admin",
                "role": {
                    "title": "Engine_Admin",
                    "includedPermissions": ADMIN_PERMISSIONS,
                    "stage": "GA"
                }
            }
        )
        if response.status_code != 200 and 'ALREADY_EXISTS' not in response.text:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Invalid response from Google API")
        print("POST", roles_url)
        response = user_session.post(
            roles_url,
            headers={
                'Content-Type': 'application/json'
            },
            json={
                "roleId": "Lapdog_user",
                "role": {
                    "title": "Lapdog_user",
                    "includedPermissions": [
                        "genomics.operations.get",
                    ],
                    "stage": "GA"
                }
            }
        )
        if response.status_code != 200 and 'ALREADY_EXISTS' not in response.text:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Invalid response from Google API")

        print("Creating Core Service Account")
        cmd = (
            'gcloud --project {project} iam service-accounts create lapdog-worker --display-name lapdog-worker'.format(
                project=custom_lapdog_project
            )
        )
        print(cmd)
        result = run_cmd(cmd)
        if result.returncode != 0 and b'already exists' not in result.buffer:
            raise ValueError("Unable to create service account")
        core_account = 'lapdog-worker@{}.iam.gserviceaccount.com'.format(custom_lapdog_project)
        print("Creating Cloud Functions Service Account")
        cmd = (
            'gcloud --project {project} iam service-accounts create lapdog-functions --display-name lapdog-functions'.format(
                project=custom_lapdog_project
            )
        )
        print(cmd)
        result = run_cmd(cmd)
        if result.returncode != 0 and b'already exists' not in result.buffer:
            raise ValueError("Unable to create service account")
        functions_account = 'lapdog-functions@{}.iam.gserviceaccount.com'.format(custom_lapdog_project)
        print("Creating Self-Updater Service Account")
        cmd = (
            'gcloud --project {project} iam service-accounts create lapdog-update --display-name lapdog-update'.format(
                project=custom_lapdog_project
            )
        )
        print(cmd)
        result = run_cmd(cmd)
        if result.returncode != 0 and b'already exists' not in result.buffer:
            raise ValueError("Unable to create service account")
        update_account = 'lapdog-update@{}.iam.gserviceaccount.com'.format(custom_lapdog_project)
        iam_url = 'https://iam.googleapis.com/v1/projects/{project}/serviceAccounts/{account}:setIamPolicy'.format(
            project=custom_lapdog_project,
            account=update_account
        )
        print("POST", iam_url)
        response = user_session.post(
            iam_url,
            headers={'Content-Type': 'application/json'},
            json={
                "policy": {
                    "bindings": [
                        {
                            "role": "roles/iam.serviceAccountUser",
                            "members": [
                                # Allows the gcloud functions account to set this pet account on comwell servers
                                "serviceAccount:{email}".format(email=functions_account),
                            ]
                        }
                    ]
                },
                "updateMask": "bindings"
            }
        )
        if response.status_code != 200:
            raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
        print("Creating Metadata bucket while service accounts are created")
        cmd = (
            'gsutil mb -c Standard -l us-central1 -p {project} gs://{bucket}'.format(
                project=custom_lapdog_project,
                bucket=ld_meta_bucket_for_project(custom_lapdog_project)
            )
        )
        print(cmd)
        result = run_cmd(cmd)
        if result.returncode != 0 and b'already exists' not in result.buffer:
            raise ValueError("Unable to create bucket")
        print("Updating project IAM policy while service accounts are created")
        policy = {
            'serviceAccount:'+core_account: 'Core_account',
            'serviceAccount:'+functions_account: 'Functions_account',
            'serviceAccount:'+update_account: 'Engine_Admin'
        }
        print(policy)
        status, response = update_iam_policy(
            user_session,
            policy,
            custom_lapdog_project
        )
        if not status:
            print('(%d) : %s' % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Invalid Response from Google API")
        print("Waiting for service account creation...")
        time.sleep(30)
        print("Issuing Core Service Account Key")
        _generate_core_key_internal(custom_lapdog_project, functions_account, user_session)
        print("Saving Project Resolution to Engine")
        blob = getblob(
            'gs://{bucket}/resolution'.format(
                bucket=ld_meta_bucket_for_project(custom_lapdog_project)
            )
        )
        blob.upload_from_string(project_id.encode())
        acl = blob.acl
        acl.all_authenticated().grant_read()
        acl.save()
        print("Configuring VPC Subnet")
        subnet_url = "https://www.googleapis.com/compute/v1/projects/{project}/regions/us-central1/subnetworks/default/setPrivateIpGoogleAccess".format(
            project=custom_lapdog_project
        )
        print("POST", subnet_url)
        response = user_session.post(
            subnet_url,
            headers={
                'Content-Type': "application/json"
            },
            params={
                'requestId': str(uuid4())
            },
            json={
                "privateIpGoogleAccess": True
            }
        )
        if response.status_code >= 400:
            raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
        print("Deploying Cloud Functions")
        from .cloud import _deploy
        _deploy('update', 'update', functions_account, custom_lapdog_project)
        _deploy('create_submission', 'submit', functions_account, custom_lapdog_project)
        _deploy('abort_submission', 'abort', functions_account, custom_lapdog_project)
        _deploy('check_abort', 'signature', functions_account, custom_lapdog_project)
        _deploy('register', 'register', functions_account, custom_lapdog_project)
        _deploy('query_account', 'query', functions_account, custom_lapdog_project)
        _deploy('quotas', 'quotas', functions_account, custom_lapdog_project)
        # Important that existence is deployed last
        # Once deployed, lapdog gateways will start reporting that the Engine is active
        _deploy('existence', 'existence', functions_account, custom_lapdog_project)

    @property
    def registered(self):
        """
        Property. Verifies that the current user is registered with this gateway
        """
        response = get_user_session().post(
            self.get_endpoint('query'),
        )
        return response.status_code == 200

    def register(self, workspace, bucket):
        """
        Registers the currently logged in account with the api
        Must provide a workspace name and the associated bucket to prove that you have access
        to at least one workspace for this namespace
        """
        warnings.warn("[BETA] Gateway Register")
        response = get_user_session().post(
            self.get_endpoint('register'),
            headers={'Content-Type': 'application/json'},
            json={
                'bucket': bucket,
                'namespace': self.namespace,
                'workspace': workspace,
            }
        )
        if response.status_code != 200:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Gateway failed to register user")
        return response.text # your account email

    def create_submission(self, workspace, bucket, submission_id, workflow_options=None, use_cache=True, memory=3, private=False, region=None, _cache_size=None):
        """
        Sends a request through the lapdog execution API to start a new submission.
        Takes the local submission ID.
        Assumes the following files to be in place:
        Workflow Inputs : gs://{workspace bucket}/lapdog-executions/{submission id}/config.json
        Workflow WDL : gs://{workspace bucket}/lapdog-executions/{submission id}/method.wdl
        Submission JSON: gs://{workspace bucket}/lapdog-executions/{submission id}/submission.json

        The user will already have needed access to the workspace in order to evaluate
        workflow inputs.
        The user must have been granted access to the lapdog execution API for this
        workspace's namespace in order for the API to accept the request.
        The user's service account must have access to the workspace in order to
        download the input files specified in the workflow inputs.
        """
        warnings.warn("[BETA] Gateway Create Submission")
        if _cache_size is None and use_cache:
            blob = getblob('gs://{}/lapdog-call-cache.sql'.format(bucket))
            if blob.exists():
                blob.reload()
                _cache_size = blob.size
        response = get_user_session().post(
            self.get_endpoint('submit'),
            headers={'Content-Type': 'application/json'},
            json={
                'bucket': bucket,
                'submission_id': submission_id,
                'namespace': self.namespace,
                'workspace': workspace,
                'workflow_options': workflow_options if workflow_options is not None else {},
                'memory': memory*1024,
                'no_ip': private,
                'compute_region': region,
                'callcache': use_cache,
                'cache_size': _cache_size / 1073741824 # 1gib
            }
        )
        if response.status_code == 200:
            operation = response.text
            submission_data_path = 'gs://{bucket}/lapdog-executions/{submission_id}/submission.json'.format(
                bucket=bucket,
                submission_id=submission_id
            )
            blob = getblob(submission_data_path)

            blob.upload_from_string(
                json.dumps(
                    {
                        **json.loads(blob.download_as_string().decode()),
                        **{'operation': operation}
                    }
                ).encode()
            )
            cache_write(
                "{}/{}/{}".format(
                    self.namespace,
                    workspace,
                    submission_id
                ),
                'submission-pointer',
                bucket,
                submission_id
            )
            return True, operation
        return False, response


    def abort_submission(self, bucket, submission_id, hard=False):
        """
        Sends a request through the lapdog execution API to abort a running submission.
        Takes the local submission ID and a list of operations corresponding to
        the workflow calls
        Assumes the following file to be in place:
        Submission JSON: gs://{workspace bucket}/lapdog-executions/{submission id}/submission.json

        Cancels the cromwell server operation then cancels all workflow operations,
        then deletes all workflow machines, then finally the cromwell machine.
        """
        warnings.warn("[BETA] Gateway Abort Submission")
        response = get_user_session().delete(
            self.get_endpoint('abort'),
            headers={'Content-Type': 'application/json'},
            json={
                'bucket': bucket,
                'submission_id': submission_id,
                'hard': hard
            }
        )
        if response.status_code != 200:
            return response

    def get_endpoint(self, endpoint, _version=None):
        """
        1) Generates the appropriate url for a given endpoint in this project
        2) Checks that the endpoint exists by submitting an OPTIONS request
        3) Returns the full endpoint url
        """
        if self.project is None:
            raise ValueError("No resolution for namespace %s. Project may not be initialized. Please contact the namespace admin" % self.namespace)
        if _version is None:
            if endpoint not in __API_VERSION__:
                raise KeyError("Endpoint not defined: "+endpoint)
            _version = __API_VERSION__[endpoint]
        endpoint_url = 'https://us-central1-{project}.cloudfunctions.net/{endpoint}-{version}'.format(
            project=self.project,
            endpoint=quote(endpoint),
            version=_version
        )
        response = get_user_session().options(endpoint_url)
        if response.status_code == 204:
            return endpoint_url
        if response.status_code == 200 or response.status_code == 404:
            print("Lapdog Engine Project", self.project, "for namespace", self.namespace, "does not support api version", _version, file=sys.stderr)
            if endpoint =='existence':
                raise ValueError("The existence endpoint could not be found. Project %s may not be initialized. Please contact the namespace admin" % self.project)
            raise ValueError("The project api for %s does not support %s version %s. Please contact the namespace admin" % (
                self.project,
                endpoint,
                _version
            ))
        raise ValueError("Unexpected status (%d) when checking for endpoint" % response.status_code)

    @property
    def exists(self):
        """
        Property. Checks that the current Gateway actually exists.
        Checks that a specific internal endpoint exists and returns the expected response
        """
        try:
            response = get_user_session().get(self.get_endpoint('existence'))
            return response.status_code == 200 and response.text == 'OK'
        except ValueError:
            return False

    @property
    def quota_usage(self):
        """
        Property. Connects to the Gateway to fetch the current quota usage
        """
        warnings.warn("[BETA] Gateway Quotas")
        response = get_user_session().post(
            self.get_endpoint('quotas'),
        )
        if response.status_code == 200:
            return response.json()
        print("Quota error (%d) : %s" % (response.status_code, response.text), file=sys.stderr)
        raise ValueError("Unable to fetch quotas: status %d" % response.status_code)

    @property
    def compute_regions(self):
        if self.project is None:
            raise ValueError("Unable to check compute regions without a working Engine")
        return enabled_regions(self.project)

    @compute_regions.setter
    def compute_regions(self, regions):
        """
        Sets the list of allowed compute regions for this gateway.
        You must have Editor permissions to the project for this namespace
        """
        if self.project is None:
            raise ValueError("Unable to set compute regions without a working Engine")
        if len(regions) <= 0:
            raise ValueError("Must provide at least one compute region")
        user_session = get_user_session()
        print("Checking VPC configuration for new regions")
        for region in regions:
            if region not in GCP_ZONES:
                raise NameError(region + " is not a valid GCP Region")
            subnet_url = "https://www.googleapis.com/compute/v1/projects/{project}/regions/{region}/subnetworks/default".format(
                project=self.project,
                region=region
            )
            response = user_session.get(subnet_url)
            if response.status_code != 200:
                raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
            subnet = response.json()
            if not ('privateIpGoogleAccess' in subnet and subnet['privateIpGoogleAccess']):
                print("Updating VPC Subnet configuration for", region)
                response = user_session.post(
                    subnet_url+'/setPrivateIpGoogleAccess',
                    headers={
                        'Content-Type': "application/json"
                    },
                    params={
                        'requestId': str(uuid4())
                    },
                    json={
                        "privateIpGoogleAccess": True
                    }
                )
                if response.status_code >= 400:
                    raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
        blob = getblob('gs://{bucket}/regions'.format(bucket=ld_meta_bucket_for_project(self.project)))
        blob.upload_from_string("\n".join(regions))
        acl = blob.acl
        acl.all_authenticated().grant_read()
        acl.save()

    def __repr__(self):
        return '<lapdog.Gateway {}{}>'.format(
            self.namespace,
            ' ({})'.format(self.project) if self.project is not None else ''
        )
